import webapp2
from tests.base_test import BaseTest

import json
import webtest
import graphene
from graphene import relay
from graphene_gae.webapp2 import graphql_application, GraphQLHandler

__author__ = 'ekampf'


class QueryRootType(graphene.ObjectType):
    default_greet = 'World'

    greet = graphene.Field(graphene.String, who=graphene.Argument(graphene.String))
    resolver_raises = graphene.String()

    @graphene.resolve_only_args
    def resolve_greet(self, who):
        return 'Hello %s!' % who

    def resolve_resolver_raises(self, *_):
        raise Exception("TEST")


class ChangeDefaultGreetingMutation(relay.ClientIDMutation):
    class Input:
        value = graphene.String()

    ok = graphene.Boolean()
    defaultGreeting = graphene.String()

    @classmethod
    def mutate_and_get_payload(cls, input, context, info):
        QueryRootType.default_greet = input.get('value')
        return ChangeDefaultGreetingMutation(ok=True, defaultGreeting=QueryRootType.default_greet)


class MutationRootType(graphene.ObjectType):
    changeDefaultGreeting = ChangeDefaultGreetingMutation.Field()


schema = graphene.Schema(query=QueryRootType, mutation=MutationRootType)

graphql_application.config['graphql_schema'] = schema
graphql_application.config['graphql_pretty'] = True


class TestGraphQLHandler(BaseTest):
    def setUp(self):
        BaseTest.setUp(self)

        self.app = webtest.TestApp(graphql_application)

    def testPOST_noSchema_returns500(self):
        graphql_application = webapp2.WSGIApplication([
            ('/graphql', GraphQLHandler)
        ])

        app = webtest.TestApp(graphql_application)
        response = app.post('/graphql', expect_errors=True)
        self.assertEqual(response.status_int, 500)
        self.assertEqual(response.json_body['errors'][0]['message'], 'GraphQL Schema is missing.')

    def testPOST_noInput_returns400(self):
        response = self.app.post('/graphql', expect_errors=True)
        self.assertEqual(response.status_int, 400)

    def testPost_supports_operation_name(self):
        response = self.app.post('/graphql', params=json.dumps(dict(
            query='''
                query helloYou { greet(who: "You"), ...shared }
                query helloWorld { greet(who: "World"), ...shared }
                query helloDolly { greet(who: "Dolly"), ...shared }
                fragment shared on QueryRootType {
                    shared: greet(who: "Everyone")
                }
            ''',
            operation_name='helloDolly'
        )))

        response_dict = json.loads(response.body)
        self.assertDictEqual(
            response_dict.get('data'),
            {
                'greet': 'Hello Dolly!',
                'shared': 'Hello Everyone!'
            }
        )

    def testPOST_support_json_variables(self):
        response = self.app.post('/graphql', params=json.dumps(dict(
            query='query helloWho($who: String){ greet(who: $who) }',
            variables={'who': "ekampf"}
        )))

        response_dict = json.loads(response.body)
        self.assertDictEqual(
            response_dict.get('data'), {'greet': 'Hello ekampf!'}
        )

    def testPOST_reports_argument_validation_errors(self):
        response = self.app.post('/graphql', expect_errors=True, params=json.dumps(dict(
            query='''
                query helloYou { greet(who: 123), ...shared }
                query helloWorld { greet(who: "World"), ...shared }
                query helloDolly { greet(who: "Dolly"), ...shared }
                fragment shared on Query {
                    shared: greet(who: "Everyone")
                }
            ''',
            operation_name='helloYou'
        )))

        self.assertEqual(response.status_int, 400)

        response_dict = json.loads(response.body)
        self.assertEqual(response_dict["errors"][0]["message"], "Argument \"who\" has invalid value 123.\nExpected type \"String\", found 123.")

    def testPOST_reports_missing_operation_name(self):
        response = self.app.post('/graphql', expect_errors=True, params=json.dumps(dict(
            query='''
                query helloWorld { greet(who: "World"), ...shared }
                query helloDolly { greet(who: "Dolly"), ...shared }
                fragment shared on QueryRootType {
                    shared: greet(who: "Everyone")
                }
            '''
        )))

        self.assertEqual(response.status_int, 400)

        response_dict = json.loads(response.body)
        self.assertEqual(response_dict["errors"][0]["message"], "Must provide operation name if query contains multiple operations.")

    def testPOST_handles_syntax_errors(self):
        response = self.app.post('/graphql', expect_errors=True, params=json.dumps(dict(query='syntaxerror')))

        self.assertEqual(response.status_int, 400)

        expected = {
            'errors': [{'locations': [{'column': 1, 'line': 1}],
                        'message': 'Syntax Error GraphQL request (1:1) '
                                   'Unexpected Name "syntaxerror"\n\n1: syntaxerror\n   ^\n'}]
        }
        response_dict = json.loads(response.body)
        self.assertEqual(response_dict, expected)

    def testPOST_handles_poorly_formed_variables(self):
        response = self.app.post('/graphql', expect_errors=True, params=json.dumps(dict(
            query='query helloWho($who: String){ greet(who: $who) }',
            variables='who:You'
        )))

        response_data = json.loads(response.body)
        self.assertEqual(response.status_int, 400)
        self.assertEqual(response_data['errors'][0]['message'], 'Variables are invalid JSON.')

    def testPOST_mutations(self):
        response = self.app.post('/graphql',
                                 json.dumps(
                                     dict(
                                         query='''
                                         mutation TestMutatio  {
                                            changeDefaultGreeting(input: { value: "universe" } ) {
                                                ok,
                                                defaultGreeting
                                            }
                                        }'''
                                     )
                                 ))

        self.assertEqual('universe', response.json_body['data']['changeDefaultGreeting']['defaultGreeting'])
        self.assertEqual(True, response.json_body['data']['changeDefaultGreeting']['ok'])

    def testPOST_override_pretty_via_query_param(self):
        response = self.app.post('/graphql?pretty=true', params=json.dumps(dict(
            query='query helloYou { greet(who: "You") }'
        )))
        self.assertEqual(response.body, '{\n  "data": {\n    "greet": "Hello You!"\n  }\n}')

    def testPOST_override_pretty_via_post_param(self):
        response = self.app.post('/graphql', params=json.dumps(dict(
            query='query helloYou { greet(who: "You") }',
            pretty=True
        )))
        self.assertEqual(response.body, '{\n  "data": {\n    "greet": "Hello You!"\n  }\n}')

    def testPOST_stringBody_readsQueryFromBodyAndRestFromGET(self):
        response = self.app.post('/graphql?pretty=True', params='query helloYou { greet(who: "You") }')
        self.assertEqual(response.body, '{\n  "data": {\n    "greet": "Hello You!"\n  }\n}')




from graphene import resolve_only_args

from tests.base_test import BaseTest

from google.appengine.ext import ndb

import graphene
from graphene.relay import Node
from graphene_gae import NdbObjectType
from graphene_gae.ndb.fields import NdbConnectionField

from tests.models import Tag, Comment, Article, Author, Address, PhoneNumber, Reader, ArticleReader

__author__ = 'ekampf'


class AddressType(NdbObjectType):
    class Meta:
        model = Address
        interfaces = (Node,)


class PhoneNumberType(NdbObjectType):
    class Meta:
        model = PhoneNumber
        interfaces = (Node,)


class ReaderType(NdbObjectType):
    class Meta:
        model = Reader
        interfaces = (Node,)


class AuthorType(NdbObjectType):
    class Meta:
        model = Author
        interfaces = (Node,)


class TagType(NdbObjectType):
    class Meta:
        model = Tag
        interfaces = (Node,)


class CommentType(NdbObjectType):
    class Meta:
        model = Comment
        interfaces = (Node,)


def transform_to_reader_edges(edges, args, context):
    article_readers = [edge.node for edge in edges]
    readers = ndb.get_multi([article_reader.reader_key for article_reader in article_readers])
    transformed_edges = []
    for edge, reader in zip(edges, readers):
        if reader.is_alive:
            edge.node = reader
            transformed_edges.append(edge)

    return transformed_edges


def reader_filter(reader, args, context):
    return reader.is_alive


class ArticleType(NdbObjectType):
    class Meta:
        model = Article
        interfaces = (Node,)

    comments = NdbConnectionField(CommentType)
    readers = NdbConnectionField(ReaderType, transform_edges=transform_to_reader_edges)

    @resolve_only_args
    def resolve_comments(self):
        return Comment.query(ancestor=self.key)

    def resolve_readers(self, args, context, info):
        return ArticleReader.query().filter(ArticleReader.article_key == self.key).order(ArticleReader.key)


class QueryRoot(graphene.ObjectType):
    articles = NdbConnectionField(ArticleType)


schema = graphene.Schema(query=QueryRoot)


class TestNDBTypesRelay(BaseTest):

    def testNdbNode_getNode_invalidId_shouldReturnNone(self):
        result = ArticleType.get_node("I'm not a valid NDB encoded key")
        self.assertIsNone(result)

    def testNdbNode_getNode_validID_entityDoesntExist_shouldReturnNone(self):
        article_key = ndb.Key('Article', 'invalid_id_thats_not_in_db')
        result = ArticleType.get_node(article_key.urlsafe())
        self.assertIsNone(result)

    def testNdbNode_getNode_validID_entityDoes_shouldReturnEntity(self):
        article_key = Article(
            headline="TestGetNode",
            summary="1",
            author_key=Author(name="John Dow", email="john@dow.com").put(),
        ).put()

        result = ArticleType.get_node(article_key.urlsafe())
        article = article_key.get()

        self.assertIsNotNone(result)
        self.assertEqual(result.headline, article.headline)
        self.assertEqual(result.summary, article.summary)
        # self.assertEqual(result.author_key, article_key.author_key)  # TODO

    def test_keyProperty(self):
        Article(
            headline="Test1",
            summary="1",
            author_key=Author(name="John Dow", email="john@dow.com").put(),
            tags=[
                Tag(name="tag1").put(),
                Tag(name="tag2").put(),
                Tag(name="tag3").put(),
            ]
        ).put()

        result = schema.execute("""
            query Articles {
                articles(first:2) {
                    edges {
                        cursor,
                        node {
                            headline,
                            summary,
                            author { name },
                            tags { name }
                        }
                    }

                }
            }
            """)

        self.assertEmpty(result.errors, msg=str(result.errors))

        articles = result.data.get('articles', {}).get('edges', [])
        self.assertLength(articles, 1)

        article = articles[0]['node']
        self.assertEqual(article['headline'], 'Test1')
        self.assertEqual(article['summary'], '1')

        author = article['author']
        self.assertLength(author.keys(), 1)
        self.assertEqual(author['name'], 'John Dow')

        tags = article['tags']
        tag_names = [t['name'] for t in tags]
        self.assertListEqual(tag_names, ['tag1', 'tag2', 'tag3'])

    def test_connectionField(self):
        a1 = Article(headline="Test1", summary="1").put()
        a2 = Article(headline="Test2", summary="2").put()
        a3 = Article(headline="Test3", summary="3").put()

        Comment(parent=a1, body="c1").put()
        Comment(parent=a2, body="c2").put()
        Comment(parent=a3, body="c3").put()

        result = schema.execute("""
            query Articles {
                articles(first:2) {
                    edges {
                        cursor,
                        node {
                            headline,
                            summary,
                            comments {
                                edges {
                                    cursor,
                                    node {
                                        body
                                    }
                                }
                            }
                        }
                    }

                }
            }
        """)

        self.assertEmpty(result.errors)

        articles = result.data.get('articles', {}).get('edges')
        self.assertLength(articles, 2)

        for articleNode in articles:
            article = articleNode['node']
            self.assertLength(article.keys(), 3)
            self.assertIsNotNone(article.get('headline'))
            self.assertIsNotNone(article.get('summary'))

            comments = article['comments']['edges']
            self.assertLength(comments, 1)
            self.assertEqual(comments[0]['node']['body'], "c" + article['summary'])

    def test_connectionField_keysOnly(self):
        a1 = Article(headline="Test1", summary="1").put()
        a2 = Article(headline="Test2", summary="2").put()
        a3 = Article(headline="Test3", summary="3").put()

        Comment(parent=a1, body="c1").put()
        Comment(parent=a2, body="c2").put()
        Comment(parent=a3, body="c3").put()

        result = schema.execute("""
            query Articles {
                articles(keysOnly: true) {
                    edges {
                        cursor,
                        node {
                            id
                        }
                    }

                }
            }
            """)

        self.assertEmpty(result.errors)

        articles = result.data.get('articles', {}).get('edges')
        self.assertLength(articles, 3)

    def test_connectionField_empty(self):
        Article(headline="Test1", summary="1").put()

        result = schema.execute("""
            query Articles {
                articles {
                    edges {
                        cursor,
                        node {
                            headline,
                            createdAt,
                            comments {
                                edges {
                                    node {
                                        body
                                    }
                                }
                            }
                        }
                    }

                }
            }
            """)

        articles = result.data.get('articles', {}).get('edges')
        self.assertLength(articles, 1)

        article = articles[0]['node']
        self.assertLength(article.keys(), 3)
        self.assertIsNotNone(article.get('headline'))
        self.assertIsNotNone(article.get('createdAt'))

        comments = article['comments']['edges']
        self.assertEmpty(comments)

    def test_connectionField_model(self):
        self.assertEqual(NdbConnectionField(CommentType).model, Comment)

    def test_connectionFieldWithTransformEdges(self):
        alive_reader_key = Reader(name="Chuck Norris", email="gmail@chuckNorris.com").put()
        dead_reader_key = Reader(name="John Doe", email="johndoe@gmail.com", is_alive=False).put()
        article_key = Article(headline="Test1", summary="1").put()

        ArticleReader(reader_key=alive_reader_key, article_key=article_key).put()
        ArticleReader(reader_key=dead_reader_key, article_key=article_key).put()

        result = schema.execute("""
            query Articles {
                articles {
                    edges {
                        node {
                            readers {
                                edges {
                                    cursor
                                    node {
                                        name
                                        email
                                    }
                                }
                            }
                        }
                    }

                }
            }
            """)

        articles = result.data.get('articles', {}).get('edges')
        self.assertLength(articles, 1)

        article = articles[0]['node']

        readers = article.get('readers', {}).get('edges')
        self.assertLength(readers, 1)

        reader = readers[0]['node']

        self.assertLength(reader.keys(), 2)
        self.assertIsNotNone(reader.get('name'))
        self.assertIsNotNone(reader.get('email'))

    def test_connectionFieldWithTransformEdges_continualEdgeGeneration(self):
        dead1_reader_key = Reader(name="John Doe 1", email="johndoe@gmail.com", is_alive=False).put()
        dead2_reader_key = Reader(name="John Doe 2", email="johndoe@gmail.com", is_alive=False).put()
        alive_reader_key = Reader(name="Chuck Norris", email="gmail@chuckNorris.com").put()
        article_key = Article(headline="Test1", summary="1").put()

        ArticleReader(reader_key=dead1_reader_key, article_key=article_key).put()
        ArticleReader(reader_key=dead2_reader_key, article_key=article_key).put()
        ArticleReader(reader_key=alive_reader_key, article_key=article_key).put()

        result = schema.execute("""
            query Articles {
                articles {
                    edges {
                        node {
                            readers(first:1) {
                                edges {
                                    cursor
                                    node {
                                        ndbId
                                        name
                                        email
                                        isAlive
                                    }
                                }
                            }
                        }
                    }

                }
            }
            """)

        articles = result.data.get('articles', {}).get('edges')
        self.assertLength(articles, 1)

        article = articles[0]['node']

        readers = article.get('readers', {}).get('edges')
        self.assertLength(readers, 1)

        reader = readers[0]['node']

        self.assertLength(reader.keys(), 4)
        self.assertEquals(reader['ndbId'], str(alive_reader_key.id()))



import sys
import os

google_appengine_home = os.environ.get('GOOGLE_APPENGINE_HOME', '/usr/local/google_appengine')

for path in [google_appengine_home]:
    if path not in sys.path:
        sys.path[0:0] = [path]

import unittest

from google.appengine.ext import testbed
from google.appengine.datastore import datastore_stub_util
from google.appengine.ext import ndb
from google.appengine.ext.cloudstorage import cloudstorage_stub


# noinspection PyProtectedMemberDisneyStoreAvailability
class BaseTest(unittest.TestCase):
    # so that nosetests run
    context = None

    def __init__(self, method_name=''):
        super(BaseTest, self).__init__(method_name)
        self.datastore_probability = 1

    def setUp(self):
        super(BaseTest, self).setUp()

        root_path = '.'
        application_id = 'graphene-gae-test'

        # First, create an instance of the Testbed class.
        self.testbed = testbed.Testbed()
        self.testbed.activate()
        self.testbed.setup_env(app_id=application_id, overwrite=True)
        policy = datastore_stub_util.PseudoRandomHRConsistencyPolicy(probability=self.datastore_probability)
        self.testbed.init_datastore_v3_stub(root_path=root_path, consistency_policy=policy, require_indexes=True)
        self.testbed.init_app_identity_stub()
        self.testbed.init_blobstore_stub()
        self.testbed.init_memcache_stub()
        self.testbed.init_taskqueue_stub(root_path=root_path)
        self.testbed.init_urlfetch_stub()
        self.storage = cloudstorage_stub.CloudStorageStub(self.testbed.get_stub('blobstore').storage)
        self.testbed.init_mail_stub()
        self.testbed.init_user_stub()
        self.taskqueue_stub = self.testbed.get_stub(testbed.TASKQUEUE_SERVICE_NAME)

        ndb.get_context().clear_cache()
        ndb.get_context().set_cache_policy(lambda x: True)

    def tearDown(self):
        self.testbed.init_datastore_v3_stub(False)

        self.testbed.deactivate()
        self.testbed = None
        super(BaseTest, self).tearDown()

    def get_filtered_tasks(self, url=None, name=None, queue_names=None):
        return self.taskqueue_stub.get_filtered_tasks(url=url, name=name, queue_names=queue_names)

    # region Extra Assertions
    def assertEmpty(self, l, msg=None):
        self.assertEqual(0, len(list(l)), msg=msg or str(l))

    def assertLength(self, l, expectation, msg=None):
        self.assertEqual(len(l), expectation, msg)

    def assertEndsWith(self, str, expected_suffix):
        cond = str.endswith(expected_suffix)
        self.assertTrue(cond)

    def assertStartsWith(self, str, expected_prefix):
        cond = str.startswith(expected_prefix)
        self.assertTrue(cond)
        # endregion

    def assertPositive(self, n):
        self.assertTrue(n > 0)

    def assertNegative(self, n):
        self.assertTrue(n < 0)



from graphene.core.classtypes.objecttype import ObjectTypeOptions
from graphene.relay.types import Node
from graphene.relay.utils import is_node


class NdbOptions(ObjectTypeOptions):
    """
    Defines how Graphene will convert the NDB model.
    Supports the following properties:

    * model - which model to convert
    * only_fields - only convert the following property names
    * exclude_fields - exclude specified properties from conversion

    """

    VALID_ATTRS = ('model', 'only_fields', 'exclude_fields', 'remove_key_property_suffix')

    def __init__(self, *args, **kwargs):
        super(NdbOptions, self).__init__(*args, **kwargs)
        self.model = None
        self.valid_attrs += self.VALID_ATTRS
        self.only_fields = None
        self.exclude_fields = []

    def contribute_to_class(self, cls, name):
        super(NdbOptions, self).contribute_to_class(cls, name)
        if is_node(cls):
            self.exclude_fields = list(self.exclude_fields) + ['id']
            self.interfaces.append(Node)



import logging
import json
import webapp2
import six

from graphql import GraphQLError, format_error as format_graphql_error

__author__ = 'ekampf'


class GraphQLHandler(webapp2.RequestHandler):
    def post(self):
        schema = self._get_schema()
        pretty = self._get_pretty()

        if not schema:
            webapp2.abort(500, detail='GraphQL Schema is missing.')

        query, operation_name, variables, pretty_override = self._get_grapl_params()
        pretty = pretty if not pretty_override else pretty_override

        result = schema.execute(query,
                                operation_name=operation_name,
                                variable_values=variables,
                                context_value=self._get_context(),
                                root_value=self._get_root_value())

        response = {}
        if result.errors:
            response['errors'] = [self.__format_error(e) for e in result.errors]
            logging.warn("Request had errors: %s", response)
            self._handle_graphql_errors(result.errors)

        if result.invalid:
            logging.error("GraphQL request is invalid: %s", response)
            return self.failed_response(400, response, pretty=pretty)

        response['data'] = result.data
        return self.successful_response(response, pretty=pretty)

    def handle_exception(self, exception, debug):
        logging.exception(exception)

        status_code = 500
        if isinstance(exception, webapp2.HTTPException):
            status_code = exception.code

        self.failed_response(status_code, {
            'errors': [self.__format_error(exception)]
        })

    def _handle_graphql_errors(self, result):
        pass

    def _get_schema(self):
        return self.app.config.get('graphql_schema')

    def _get_pretty(self):
        return self.app.config.get('graphql_pretty', False)

    def _get_grapl_params(self):
        try:
            request_data = self.request.json_body
            if isinstance(request_data, basestring):
                request_data = dict(query=request_data)
        except:
            request_data = {}

        request_data.update(dict(self.request.GET))

        query = request_data.get('query', self.request.body)
        if not query:
            webapp2.abort(400, "Query is empty.")

        operation_name = request_data.get('operation_name')
        variables = request_data.get('variables')
        if variables and isinstance(variables, six.text_type):
            try:
                variables = json.loads(variables)
            except:
                raise webapp2.abort(400, 'Variables are invalid JSON.')

        pretty = request_data.get('pretty')

        return query, operation_name, variables, pretty

    def _get_root_value(self):
        return None

    def _get_context(self):
        return self.request

    def __format_error(self, error):
        if isinstance(error, GraphQLError):
            return format_graphql_error(error)

        return {'message': str(error)}

    def __json_encode(self, data, pretty=False):
        if pretty:
            return json.dumps(data, indent=2, sort_keys=True, separators=(',', ': '))

        return json.dumps(data)

    def successful_response(self, data, pretty=False):
        serialized_data = self.__json_encode(data, pretty=pretty)

        self.response.set_status(200, 'Success')
        self.response.md5_etag()
        self.response.content_type = 'application/json'
        self.response.out.write(serialized_data)

    def failed_response(self, error_code, data, pretty=False):
        serialized_data = self.__json_encode(data, pretty=pretty)

        self.response.set_status(error_code)
        self.response.content_type = 'application/json'
        self.response.out.write(serialized_data)


graphql_application = webapp2.WSGIApplication([
    ('/graphql', GraphQLHandler)
])



from collections import namedtuple

import inflect

from google.appengine.ext import ndb

from graphene import String, Boolean, Int, Float, List, NonNull, Field, Dynamic
from graphene.types.json import JSONString
from graphene.types.datetime import DateTime

from .fields import DynamicNdbKeyStringField, DynamicNdbKeyReferenceField

__author__ = 'ekampf'

ConversionResult = namedtuple('ConversionResult', ['name', 'field'])


p = inflect.engine()


def rreplace(s, old, new, occurrence):
    li = s.rsplit(old, occurrence)
    return new.join(li)


def convert_ndb_scalar_property(graphene_type, ndb_prop, **kwargs):
    kwargs['description'] = "%s %s property" % (ndb_prop._name, graphene_type)
    _type = graphene_type

    if ndb_prop._repeated:
        _type = List(_type)

    if ndb_prop._required:
        _type = NonNull(_type)

    return Field(_type, **kwargs)


def convert_ndb_string_property(ndb_prop):
    return convert_ndb_scalar_property(String, ndb_prop)


def convert_ndb_boolean_property(ndb_prop):
    return convert_ndb_scalar_property(Boolean, ndb_prop)


def convert_ndb_int_property(ndb_prop):
    return convert_ndb_scalar_property(Int, ndb_prop)


def convert_ndb_float_property(ndb_prop):
    return convert_ndb_scalar_property(Float, ndb_prop)


def convert_ndb_json_property(ndb_prop):
    return Field(JSONString, description=ndb_prop._name)


def convert_ndb_datetime_property(ndb_prop):
    return Field(DateTime, description=ndb_prop._name)


def convert_ndb_key_propety(ndb_key_prop):
    """
    Two conventions for handling KeyProperties:
    #1.
        Given:
            store_key = ndb.KeyProperty(...)

        Result is 2 fields:
            store_id  = graphene.String() -> resolves to store_key.urlsafe()
            store     = NdbKeyField()     -> resolves to entity

    #2.
        Given:
            store = ndb.KeyProperty(...)

        Result is 2 fields:
            store_id = graphene.String() -> resolves to store_key.urlsafe()
            store     = NdbKeyField()    -> resolves to entity

    """
    is_repeated = ndb_key_prop._repeated
    name = ndb_key_prop._code_name

    if name.endswith('_key') or name.endswith('_keys'):
        # Case #1 - name is of form 'store_key' or 'store_keys'
        string_prop_name = rreplace(name, '_key', '_id', 1)
        resolved_prop_name = name[:-4] if name.endswith('_key') else p.plural(name[:-5])
    else:
        # Case #2 - name is of form 'store'
        singular_name = p.singular_noun(name) if p.singular_noun(name) else name
        string_prop_name = singular_name + '_ids' if is_repeated else singular_name + '_id'
        resolved_prop_name = name

    return [
        ConversionResult(name=string_prop_name, field=DynamicNdbKeyStringField(ndb_key_prop)),
        ConversionResult(name=resolved_prop_name, field=DynamicNdbKeyReferenceField(ndb_key_prop))
    ]


def convert_local_structured_property(ndb_structured_property):
    is_required = ndb_structured_property._required
    is_repeated = ndb_structured_property._repeated
    model = ndb_structured_property._modelclass
    name = ndb_structured_property._code_name

    def dynamic_type():
        from .types import NdbObjectTypeMeta
        _type = NdbObjectTypeMeta.REGISTRY.get(model.__name__)
        if not _type:
            return None

        if is_repeated:
            _type = List(_type)

        if is_required:
            _type = NonNull(_type)

        return Field(_type)

    field = Dynamic(dynamic_type)
    return ConversionResult(name=name, field=field)


def convert_computed_property(ndb_computed_prop):
    return convert_ndb_scalar_property(String, ndb_computed_prop)


converters = {
    ndb.StringProperty: convert_ndb_string_property,
    ndb.TextProperty: convert_ndb_string_property,
    ndb.BooleanProperty: convert_ndb_boolean_property,
    ndb.IntegerProperty: convert_ndb_int_property,
    ndb.FloatProperty: convert_ndb_float_property,
    ndb.JsonProperty: convert_ndb_json_property,
    ndb.DateProperty: convert_ndb_datetime_property,
    ndb.DateTimeProperty: convert_ndb_datetime_property,
    ndb.KeyProperty: convert_ndb_key_propety,
    ndb.StructuredProperty: convert_local_structured_property,
    ndb.LocalStructuredProperty: convert_local_structured_property,
    ndb.ComputedProperty: convert_computed_property
}


def convert_ndb_property(prop):
    converter_func = converters.get(type(prop))
    if not converter_func:
        raise Exception("Don't know how to convert NDB field %s (%s)" % (prop._code_name, prop))

    field = converter_func(prop)
    if not field:
        raise Exception("Failed to convert NDB propeerty to a GraphQL field %s (%s)" % (prop._code_name, prop))

    if isinstance(field, (list, ConversionResult,)):
        return field

    return ConversionResult(name=prop._code_name, field=field)




import inspect
import six
from collections import OrderedDict

from google.appengine.ext import ndb

from graphene import ObjectType, Field, String
from graphene.types.objecttype import ObjectTypeMeta, merge, yank_fields_from_attrs
from graphene.utils.is_base_type import is_base_type
from graphene.types.options import Options

from graphene_gae.ndb.converter import convert_ndb_property

__author__ = 'ekampf'


class NdbObjectTypeMeta(ObjectTypeMeta):
    REGISTRY = {}  # Maps between ndb.Model to its GraphQL type

    def __new__(mcs, name, bases, attrs):
        if not is_base_type(bases, NdbObjectTypeMeta):
            return type.__new__(mcs, name, bases, attrs)

        options = Options(
            attrs.pop('Meta', None),
            name=name,
            description=attrs.pop('__doc__', None),
            model=None,
            local_fields=None,
            only_fields=(),
            exclude_fields=(),
            interfaces=(),
            registry=None
        )

        if not options.model:
            raise Exception('NdbObjectType %s must have a model in the Meta class attr' % name)

        if not inspect.isclass(options.model) or not issubclass(options.model, ndb.Model):
            raise Exception('Provided model in %s is not an NDB model' % name)

        new_cls = ObjectTypeMeta.__new__(mcs, name, bases, dict(attrs, _meta=options))
        mcs.register(new_cls)

        ndb_fields = mcs.fields_for_ndb_model(options)
        options.ndb_fields = yank_fields_from_attrs(
            ndb_fields,
            _as=Field,
        )
        options.fields = merge(
            options.interface_fields,
            options.ndb_fields,
            options.base_fields,
            options.local_fields
        )

        return new_cls

    @classmethod
    def register(mcs, object_type_meta):
        mcs.REGISTRY[object_type_meta._meta.model.__name__] = object_type_meta

    @staticmethod
    def fields_for_ndb_model(options):
        ndb_model = options.model
        only_fields = options.only_fields
        already_created_fields = {name for name, _ in options.local_fields.iteritems()}

        ndb_fields = OrderedDict()
        for prop_name, prop in ndb_model._properties.iteritems():
            name = prop._code_name

            is_not_in_only = only_fields and name not in only_fields
            is_excluded = name in options.exclude_fields or name in already_created_fields
            if is_not_in_only or is_excluded:
                continue

            results = convert_ndb_property(prop)
            if not results:
                continue

            if not isinstance(results, list):
                results = [results]

            for r in results:
                ndb_fields[r.name] = r.field

        ndb_fields['ndb_id'] = Field(String, resolver=lambda entity, *_: str(entity.key.id()))

        return ndb_fields


class NdbObjectType(six.with_metaclass(NdbObjectTypeMeta, ObjectType)):
    @classmethod
    def is_type_of(cls, root, context, info):
        if isinstance(root, cls):
            return True

        if not cls.is_valid_ndb_model(type(root)):
            raise Exception(('Received incompatible instance "{}".').format(root))

        return type(root) == cls._meta.model

    @classmethod
    def get_node(cls, urlsafe_key, *_):
        try:
            key = ndb.Key(urlsafe=urlsafe_key)
        except:
            return None

        model = cls._meta.model
        assert key.kind() == model.__name__
        return key.get()

    @classmethod
    def resolve_id(cls, entity, args, context, info):
        return entity.key.urlsafe()

    @staticmethod
    def is_valid_ndb_model(model):
        return inspect.isclass(model) and issubclass(model, ndb.Model)




from graphql_relay import to_global_id

from tests.base_test import BaseTest

import graphene

from graphene_gae import NdbObjectType
from tests.models import Tag, Comment, Article, Address, Author, PhoneNumber

__author__ = 'ekampf'


class AddressType(NdbObjectType):
    class Meta:
        model = Address


class PhoneNumberType(NdbObjectType):
    class Meta:
        model = PhoneNumber


class AuthorType(NdbObjectType):
    class Meta:
        model = Author


class TagType(NdbObjectType):
    class Meta:
        model = Tag


class CommentType(NdbObjectType):
    class Meta:
        model = Comment


class ArticleType(NdbObjectType):
    class Meta:
        model = Article
        exclude_fields = ['to_be_excluded']


class QueryRoot(graphene.ObjectType):
    articles = graphene.List(ArticleType)

    @graphene.resolve_only_args
    def resolve_articles(self):
        return Article.query()


schema = graphene.Schema(query=QueryRoot)


class TestNDBTypes(BaseTest):

    def testNdbObjectType_instanciation(self):
        instance = Article(headline="test123")
        h = ArticleType(**instance.to_dict(exclude=["tags", "author_key"]))
        self.assertEqual(instance.headline, h.headline)

    def testNdbObjectType_should_raise_if_no_model(self):
        with self.assertRaises(Exception) as context:
            class Character1(NdbObjectType):
                pass

        assert 'model in the Meta' in str(context.exception.message)

    def testNdbObjectType_should_raise_if_model_is_invalid(self):
        with self.assertRaises(Exception) as context:
            class Character2(NdbObjectType):
                class Meta:
                    model = 1

        assert 'not an NDB model' in str(context.exception.message)

    # def testNdbObjectType_keyProperty_kindDoesntExist_raisesException(self):
    #     with self.assertRaises(Exception) as context:
    #         class ArticleType(NdbObjectType):
    #             class Meta:
    #                 model = Article
    #                 only_fields = ('prop',)
    #
    #             prop = NdbKeyReferenceField('foo', 'bar')
    #
    #         class QueryType(graphene.ObjectType):
    #             articles = graphene.List(ArticleType)
    #
    #             @graphene.resolve_only_args
    #             def resolve_articles(self):
    #                 return Article.query()
    #
    #         schema = graphene.Schema(query=QueryType)
    #         schema.execute('query test {  articles { prop } }')
    #
    #     self.assertIn("Model 'bar' is not accessible by the schema.", str(context.exception.message))

    # def testNdbObjectType_keyProperty_stringRepresentation_kindDoesntExist_raisesException(self):
    #     with self.assertRaises(Exception) as context:
    #         class ArticleType(NdbObjectType):
    #             class Meta:
    #                 model = Article
    #                 only_fields = ('prop',)
    #
    #             prop = NdbKeyStringField('foo', 'bar')
    #
    #         class QueryType(graphene.ObjectType):
    #             articles = graphene.List(ArticleType)
    #
    #             @graphene.resolve_only_args
    #             def resolve_articles(self):
    #                 return Article.query()
    #
    #         schema = graphene.Schema(query=QueryType)
    #         schema.execute('query test {  articles { prop } }')
    #
    #     self.assertIn("Model 'bar' is not accessible by the schema.", str(context.exception.message))

    def testQuery_excludedField(self):
        Article(headline="h1", summary="s1").put()

        class ArticleType(NdbObjectType):
            class Meta:
                model = Article
                exclude_fields = ['summary']

        class QueryType(graphene.ObjectType):
            articles = graphene.List(ArticleType)

            @graphene.resolve_only_args
            def resolve_articles(self):
                return Article.query()

        schema = graphene.Schema(query=QueryType)
        query = '''
            query ArticlesQuery {
              articles { headline, summary }
            }
        '''

        result = schema.execute(query)

        self.assertIsNotNone(result.errors)
        self.assertTrue('Cannot query field "summary"' in result.errors[0].message)

    def testQuery_onlyFields(self):
        Article(headline="h1", summary="s1").put()

        class ArticleType(NdbObjectType):
            class Meta:
                model = Article
                only_fields = ['headline']

        class QueryType(graphene.ObjectType):
            articles = graphene.List(ArticleType)

            @graphene.resolve_only_args
            def resolve_articles(self):
                return Article.query()

        schema = graphene.Schema(query=QueryType)
        query = '''
                    query ArticlesQuery {
                      articles { headline }
                    }
                '''

        result = schema.execute(query)

        self.assertIsNotNone(result.data)
        self.assertEqual(result.data['articles'][0]['headline'], 'h1')

        query = '''
                    query ArticlesQuery {
                      articles { headline, summary }
                    }
                '''
        result = schema.execute(query)

        self.assertIsNotNone(result.errors)
        self.assertTrue('Cannot query field "summary"' in result.errors[0].message)

    def testQuery_list(self):
        Article(headline="Test1", summary="1").put()
        Article(headline="Test2", summary="2").put()
        Article(headline="Test3", summary="3").put()

        result = schema.execute("""
            query Articles {
                articles {
                    headline
                }
            }
        """)
        self.assertEmpty(result.errors)

        self.assertLength(result.data['articles'], 3)

        for article in result.data['articles']:
            self.assertLength(article.keys(), 1)
            self.assertEqual(article.keys()[0], 'headline')

    def testQuery_repeatedProperty(self):
        keywords = ["a", "b", "c"]
        a = Article(headline="Test1", keywords=keywords).put()

        result = schema.execute("""
            query Articles {
                articles {
                    headline,
                    keywords,
                    createdAt
                }
            }
        """)
        self.assertEmpty(result.errors)

        self.assertLength(result.data['articles'], 1)

        article = result.data['articles'][0]
        self.assertEqual(article["createdAt"], str(a.get().created_at.isoformat()))
        self.assertEqual(article["headline"], "Test1")
        self.assertListEqual(article["keywords"], keywords)

    def testQuery_structuredProperty(self):
        mobile = PhoneNumber(area="650", number="12345678")
        author_key = Author(name="John Dow", email="john@dow.com", mobile=mobile).put()
        Article(headline="Test1", author_key=author_key).put()

        result = schema.execute("""
            query Articles {
                articles {
                    headline,
                    authorId
                    author {
                        name
                        email
                        mobile { area, number }
                    }
                }
            }
        """)
        self.assertEmpty(result.errors, msg=str(result.errors))

        article = result.data['articles'][0]
        self.assertEqual(article["headline"], "Test1")

        author = article['author']
        self.assertEqual(author["name"], "John Dow")
        self.assertEqual(author["email"], "john@dow.com")
        self.assertDictEqual(dict(area="650", number="12345678"), dict(author["mobile"]))

    def testQuery_structuredProperty_repeated(self):
        address1 = Address(address1="address1", address2="apt 1", city="Mountain View")
        address2 = Address(address1="address2", address2="apt 2", city="Mountain View")
        author_key = Author(name="John Dow", email="john@dow.com", addresses=[address1, address2]).put()
        Article(headline="Test1", author_key=author_key).put()

        result = schema.execute("""
            query Articles {
                articles {
                    headline,
                    author {
                        name
                        email
                        addresses {
                            address1
                            address2
                            city
                        }
                    }
                }
            }
        """)
        self.assertEmpty(result.errors)

        article = result.data['articles'][0]
        self.assertEqual(article["headline"], "Test1")

        author = article['author']
        self.assertEqual(author["name"], "John Dow")
        self.assertEqual(author["email"], "john@dow.com")
        self.assertLength(author["addresses"], 2)

        addresses = [dict(d) for d in author["addresses"]]
        self.assertIn(address1.to_dict(), addresses)
        self.assertIn(address2.to_dict(), addresses)

    def testQuery_keyProperty(self):
        author_key = Author(name="john dow", email="john@dow.com").put()
        article_key = Article(headline="h1", summary="s1", author_key=author_key).put()

        result = schema.execute('''
            query ArticleWithAuthorID {
                articles {
                    ndbId
                    headline
                    authorId
                    authorNdbId: authorId(ndb: true)
                    author {
                        name, email
                    }
                }
            }
        ''')

        self.assertEmpty(result.errors)

        article = dict(result.data['articles'][0])
        self.assertEqual(article['ndbId'], str(article_key.id()))
        self.assertEqual(article['authorNdbId'], str(author_key.id()))

        author = dict(article['author'])
        self.assertDictEqual(author, {'name': u'john dow', 'email': u'john@dow.com'})
        self.assertEqual('h1', article['headline'])
        self.assertEqual(to_global_id('AuthorType', author_key.urlsafe()), article['authorId'])

    def testQuery_repeatedKeyProperty(self):
        tk1 = Tag(name="t1").put()
        tk2 = Tag(name="t2").put()
        tk3 = Tag(name="t3").put()
        tk4 = Tag(name="t4").put()
        Article(headline="h1", summary="s1", tags=[tk1, tk2, tk3, tk4]).put()

        result = schema.execute('''
            query ArticleWithAuthorID {
                articles {
                    headline
                    authorId
                    tagIds
                    tags {
                        name
                    }
                }
            }
        ''')

        self.assertEmpty(result.errors)

        article = dict(result.data['articles'][0])
        self.assertListEqual(map(lambda k: to_global_id('TagType', k.urlsafe()), [tk1, tk2, tk3, tk4]), article['tagIds'])

        self.assertLength(article['tags'], 4)
        for i in range(0, 3):
            self.assertEqual(article['tags'][i]['name'], 't%s' % (i + 1))



from functools import partial
import six

from google.appengine.ext import ndb
from google.appengine.ext.db import BadArgumentError, Timeout
from google.appengine.runtime import DeadlineExceededError

from graphql_relay import to_global_id
from graphql_relay.connection.connectiontypes import Edge
from graphene import relay, Argument, Boolean, Int, String, Field, List, NonNull, Dynamic
from graphene.relay.connection import PageInfo


__author__ = 'ekampf'


def generate_edges_page(ndb_iter, page_size, keys_only, edge_type):
    edges = []
    timeouts = 0
    while len(edges) < page_size:
        try:
            entity = ndb_iter.next()
        except StopIteration:
            break
        except Timeout:
            timeouts += 1
            if timeouts > 2:
                break

            continue
        except DeadlineExceededError:
            break

        if keys_only:
            # entity is actualy an ndb.Key and we need to create an empty entity to hold it
            entity = edge_type._meta.fields['node']._type._meta.model(key=entity)

        edges.append(edge_type(node=entity, cursor=ndb_iter.cursor_after().urlsafe()))

    return edges


def connection_from_ndb_query(query, args=None, connection_type=None, edge_type=None, pageinfo_type=None,
                              transform_edges=None, context=None, **kwargs):
    '''
    A simple function that accepts an ndb Query and used ndb QueryIterator object(https://cloud.google.com/appengine/docs/python/ndb/queries#iterators)
    to returns a connection object for use in GraphQL.
    It uses array offsets as pagination,
    so pagination will only work if the array is static.
    '''
    args = args or {}
    connection_type = connection_type or relay.Connection
    edge_type = edge_type or Edge
    pageinfo_type = pageinfo_type or PageInfo

    full_args = dict(args, **kwargs)
    first = full_args.get('first')
    after = full_args.get('after')
    has_previous_page = bool(after)
    keys_only = full_args.get('keys_only', False)
    batch_size = full_args.get('batch_size', 20)
    page_size = first if first else full_args.get('page_size', 20)
    start_cursor = ndb.Cursor(urlsafe=after) if after else None

    ndb_iter = query.iter(produce_cursors=True, start_cursor=start_cursor, batch_size=batch_size, keys_only=keys_only, projection=query.projection)

    edges = []
    while len(edges) < page_size:
        missing_edges_count = page_size - len(edges)
        edges_page = generate_edges_page(ndb_iter, missing_edges_count, keys_only, edge_type)

        edges.extend(transform_edges(edges_page, args, context) if transform_edges else edges_page)

        if len(edges_page) < missing_edges_count:
            break

    try:
        end_cursor = ndb_iter.cursor_after().urlsafe()
    except BadArgumentError:
        end_cursor = None

    # Construct the connection
    return connection_type(
        edges=edges,
        page_info=pageinfo_type(
            start_cursor=start_cursor.urlsafe() if start_cursor else '',
            end_cursor=end_cursor,
            has_previous_page=has_previous_page,
            has_next_page=ndb_iter.has_next()
        )
    )


class NdbConnectionField(relay.ConnectionField):
    def __init__(self, type, transform_edges=None, *args, **kwargs):
        super(NdbConnectionField, self).__init__(
            type,
            *args,
            keys_only=Boolean(),
            batch_size=Int(),
            page_size=Int(),
            **kwargs
        )

        self.transform_edges = transform_edges

    @property
    def model(self):
        return self.type._meta.node._meta.model

    @staticmethod
    def connection_resolver(resolver, connection, model, transform_edges, root, args, context, info):
        ndb_query = resolver(root, args, context, info)
        if ndb_query is None:
            ndb_query = model.query()

        return connection_from_ndb_query(
            ndb_query,
            args=args,
            connection_type=connection,
            edge_type=connection.Edge,
            pageinfo_type=PageInfo,
            transform_edges=transform_edges,
            context=context
        )

    def get_resolver(self, parent_resolver):
        return partial(self.connection_resolver, parent_resolver, self.type, self.model, self.transform_edges)


class DynamicNdbKeyStringField(Dynamic):
    def __init__(self, ndb_key_prop, *args, **kwargs):
        kind = ndb_key_prop._kind

        def get_type():
            from .types import NdbObjectTypeMeta
            kind_name = kind if isinstance(kind, six.string_types) else kind.__name__

            if not NdbObjectTypeMeta.REGISTRY.get(kind_name):
                return None

            global_type_name = NdbObjectTypeMeta.REGISTRY[kind_name].__name__
            return NdbKeyStringField(ndb_key_prop, global_type_name)

        super(DynamicNdbKeyStringField, self).__init__(
            get_type,
            *args, **kwargs
        )


class DynamicNdbKeyReferenceField(Dynamic):
    def __init__(self, ndb_key_prop, *args, **kwargs):
        kind = ndb_key_prop._kind

        def get_type():
            from .types import NdbObjectTypeMeta
            kind_name = kind if isinstance(kind, six.string_types) else kind.__name__
            _type = NdbObjectTypeMeta.REGISTRY.get(kind_name)
            if not _type:
                return None

            return NdbKeyReferenceField(ndb_key_prop, _type)

        super(DynamicNdbKeyReferenceField, self).__init__(
            get_type,
            *args, **kwargs
        )


class NdbKeyStringField(Field):
    def __init__(self, ndb_key_prop, graphql_type_name, *args, **kwargs):
        self.__ndb_key_prop = ndb_key_prop
        self.__graphql_type_name = graphql_type_name
        is_repeated = ndb_key_prop._repeated
        is_required = ndb_key_prop._required

        _type = String
        if is_repeated:
            _type = List(_type)

        if is_required:
            _type = NonNull(_type)

        kwargs['args'] = {
            'ndb': Argument(Boolean, False, description="Return an NDB id (key.id()) instead of a GraphQL global id")
        }

        super(NdbKeyStringField, self).__init__(_type, *args, **kwargs)

    def resolve_key_to_string(self, entity, args, context, info):
        is_global_id = not args.get('ndb', False)
        key_value = self.__ndb_key_prop._get_user_value(entity)
        if not key_value:
            return None

        if isinstance(key_value, list):
            return [to_global_id(self.__graphql_type_name, k.urlsafe()) for k in key_value] if is_global_id else [k.id() for k in key_value]

        return to_global_id(self.__graphql_type_name, key_value.urlsafe()) if is_global_id else key_value.id()

    def get_resolver(self, parent_resolver):
        return self.resolve_key_to_string


class NdbKeyReferenceField(Field):
    def __init__(self, ndb_key_prop, graphql_type, *args, **kwargs):
        self.__ndb_key_prop = ndb_key_prop
        self.__graphql_type = graphql_type
        is_repeated = ndb_key_prop._repeated
        is_required = ndb_key_prop._required

        _type = self.__graphql_type
        if is_repeated:
            _type = List(_type)

        if is_required:
            _type = NonNull(_type)

        super(NdbKeyReferenceField, self).__init__(_type, *args, **kwargs)

    def resolve_key_reference(self, entity, args, context, info):
        key_value = self.__ndb_key_prop._get_user_value(entity)
        if not key_value:
            return None

        if isinstance(key_value, list):
            return ndb.get_multi(key_value)

        return key_value.get()

    def get_resolver(self, parent_resolver):
        return self.resolve_key_reference





#!/usr/bin/env python
# -*- coding: utf-8 -*-

import codecs
import os
import re
import sys
import unittest
from setuptools import setup, find_packages

root_dir = os.path.abspath(os.path.dirname(__file__))


def get_build_number():
    fname = 'build.info'
    if os.path.isfile(fname):
        with open(fname) as f:
            build_number = f.read()
            build_number = re.sub("[^a-z0-9]+","", build_number, flags=re.IGNORECASE)
            return '.' + build_number

    return ''


def get_version(package_name):
    build_number = get_build_number()

    version_re = re.compile(r"^__version__ = [\"']([\w_.-]+)[\"']$")
    package_components = package_name.split('.')
    init_path = os.path.join(root_dir, *(package_components + ['__init__.py']))
    with codecs.open(init_path, 'r', 'utf-8') as f:
        for line in f:
            match = version_re.match(line[:-1])
            if match:
                return match.groups()[0]+build_number

    return '0.1.0' + build_number

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read().replace('.. :changelog:', '')


requirements = [
    'six>=1.10.0',
    'inflect==0.2.5',
    'graphene>=1.0',
    'iso8601'
]

test_requirements = [
    'PyYAML==3.11',
    'webapp2==2.5.2',
    'webob==1.2.3',
    'WebTest==2.0.11',
    'mock'
]


def additional_tests():
    setup_file = sys.modules['__main__'].__file__
    setup_dir = os.path.abspath(os.path.dirname(setup_file))
    print("*** Looking for tests in %s" % setup_dir)
    return unittest.defaultTestLoader.discover(setup_dir)


setup(
    name='graphene_gae',
    version=get_version('graphene_gae'),
    description="Graphene GAE Integration",
    long_description=readme + '\n\n' + history,
    author="Eran Kampf",
    author_email='eran@ekampf.com',
    url='https://github.com/graphql-python/graphene-gae',
    packages=find_packages(exclude=['tests*', 'examples*']),
    include_package_data=True,
    install_requires=requirements,
    license="BSD",
    zip_safe=False,
    keywords='graphene_gae',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
    ],
    test_suite='setup.additional_tests',
    tests_require=test_requirements
)



# -*- coding: utf-8 -*-

from .ndb.types import (
    NdbObjectType
)

from .ndb.fields import (
    NdbConnectionField,
)

__author__ = 'Eran Kampf'
__version__ = '1.0.6'

__all__ = [
    NdbObjectType,
    NdbConnectionField,
]



import mock

from graphene_gae.ndb.types import NdbObjectType
from tests.base_test import BaseTest

from google.appengine.ext import ndb

import graphene
from graphene import List, NonNull, String
from graphene.types.json import JSONString
from graphene.types.datetime import DateTime

from graphene_gae.ndb.fields import NdbKeyStringField, NdbKeyReferenceField, DynamicNdbKeyStringField, DynamicNdbKeyReferenceField
from graphene_gae.ndb.converter import convert_ndb_property

__author__ = 'ekampf'


class SomeWeirdUnknownProperty(ndb.Property):
    pass


class TestNDBConverter(BaseTest):
    def __assert_conversion(self, ndb_property_type, expected_graphene_type, *args, **kwargs):
        ndb_property = ndb_property_type(*args, **kwargs)
        result = convert_ndb_property(ndb_property)
        graphene_field = result.field
        self.assertEqual(graphene_field._type, expected_graphene_type)

    def testUnknownProperty_raisesException(self):
        with self.assertRaises(Exception) as context:
            prop = SomeWeirdUnknownProperty()
            prop._code_name = "my_prop"
            convert_ndb_property(prop)

        self.assertTrue("Don't know how to convert" in context.exception.message, msg=context.exception.message)

    @mock.patch('graphene_gae.ndb.converter.converters')
    def testNoneResult_raisesException(self, patch_convert):
        from graphene_gae.ndb.converter import convert_ndb_property
        patch_convert.get.return_value = lambda _: None
        with self.assertRaises(Exception) as context:
            prop = ndb.StringProperty()
            prop._code_name = "my_prop"
            convert_ndb_property(prop)

        expected_message = 'Failed to convert NDB propeerty to a GraphQL field my_prop (StringProperty())'
        self.assertTrue(expected_message in context.exception.message, msg=context.exception.message)

    def testStringProperty_shouldConvertToString(self):
        self.__assert_conversion(ndb.StringProperty, graphene.String)

    def testStringProperty_repeated_shouldConvertToList(self):
        ndb_prop = ndb.StringProperty(repeated=True)
        result = convert_ndb_property(ndb_prop)
        graphene_type = result.field._type

        self.assertIsInstance(graphene_type, graphene.List)
        self.assertEqual(graphene_type.of_type, graphene.String)

    def testStringProperty_required_shouldConvertToList(self):
        ndb_prop = ndb.StringProperty(required=True)
        result = convert_ndb_property(ndb_prop)
        graphene_type = result.field._type

        self.assertIsInstance(graphene_type, graphene.NonNull)
        self.assertEqual(graphene_type.of_type, graphene.String)

    def testTextProperty_shouldConvertToString(self):
        self.__assert_conversion(ndb.TextProperty, graphene.String)

    def testBoolProperty_shouldConvertToString(self):
        self.__assert_conversion(ndb.BooleanProperty, graphene.Boolean)

    def testIntProperty_shouldConvertToString(self):
        self.__assert_conversion(ndb.IntegerProperty, graphene.Int)

    def testFloatProperty_shouldConvertToString(self):
        self.__assert_conversion(ndb.FloatProperty, graphene.Float)

    def testDateProperty_shouldConvertToString(self):
        self.__assert_conversion(ndb.DateProperty, DateTime)

    def testDateTimeProperty_shouldConvertToString(self):
        self.__assert_conversion(ndb.DateTimeProperty, DateTime)

    def testJsonProperty_shouldConvertToString(self):
        self.__assert_conversion(ndb.JsonProperty, JSONString)

    def testKeyProperty_withSuffix(self):
        class User(ndb.Model):
            name = ndb.StringProperty()

        class UserType(NdbObjectType):
            class Meta:
                model = User

        prop = ndb.KeyProperty(kind='User')
        prop._code_name = 'user_key'

        conversion = convert_ndb_property(prop)

        self.assertLength(conversion, 2)

        self.assertEqual(conversion[0].name, 'user_id')
        self.assertIsInstance(conversion[0].field, DynamicNdbKeyStringField)
        _type = conversion[0].field.get_type()
        self.assertIsInstance(_type, NdbKeyStringField)
        self.assertEqual(_type._type, String)

        self.assertEqual(conversion[1].name, 'user')
        self.assertIsInstance(conversion[1].field, DynamicNdbKeyReferenceField)
        _type = conversion[1].field.get_type()
        self.assertIsInstance(_type, NdbKeyReferenceField)
        self.assertEqual(_type._type, UserType)

    def testKeyProperty_withSuffix_repeated(self):
        class User(ndb.Model):
            name = ndb.StringProperty()

        class UserType(NdbObjectType):
            class Meta:
                model = User

        prop = ndb.KeyProperty(kind='User', repeated=True)
        prop._code_name = 'user_keys'

        conversion = convert_ndb_property(prop)

        self.assertLength(conversion, 2)

        self.assertEqual(conversion[0].name, 'user_ids')
        self.assertIsInstance(conversion[0].field, DynamicNdbKeyStringField)
        _type = conversion[0].field.get_type()
        self.assertIsInstance(_type, NdbKeyStringField)
        self.assertIsInstance(_type._type, List)
        self.assertEqual(_type._type.of_type, String)

        self.assertEqual(conversion[1].name, 'users')
        self.assertIsInstance(conversion[1].field, DynamicNdbKeyReferenceField)
        _type = conversion[1].field.get_type()
        self.assertIsInstance(_type, NdbKeyReferenceField)
        self.assertIsInstance(_type._type, List)
        self.assertEqual(_type._type.of_type, UserType)

    def testKeyProperty_withSuffix_required(self):
        class User(ndb.Model):
            name = ndb.StringProperty()

        class UserType(NdbObjectType):
            class Meta:
                model = User

        prop = ndb.KeyProperty(kind='User', required=True)
        prop._code_name = 'user_key'

        conversion = convert_ndb_property(prop)

        self.assertLength(conversion, 2)

        self.assertEqual(conversion[0].name, 'user_id')
        self.assertIsInstance(conversion[0].field, DynamicNdbKeyStringField)
        _type = conversion[0].field.get_type()
        self.assertIsInstance(_type, NdbKeyStringField)
        self.assertIsInstance(_type._type, NonNull)
        self.assertEqual(_type._type.of_type, String)

        self.assertEqual(conversion[1].name, 'user')
        self.assertIsInstance(conversion[1].field, DynamicNdbKeyReferenceField)
        _type = conversion[1].field.get_type()
        self.assertIsInstance(_type, NdbKeyReferenceField)
        self.assertIsInstance(_type._type, NonNull)
        self.assertEqual(_type._type.of_type, UserType)

    def testKeyProperty_withoutSuffix(self):
        class User(ndb.Model):
            name = ndb.StringProperty()

        class UserType(NdbObjectType):
            class Meta:
                model = User

        prop = ndb.KeyProperty(kind='User')
        prop._code_name = 'user'

        conversion = convert_ndb_property(prop)

        self.assertLength(conversion, 2)

        self.assertEqual(conversion[0].name, 'user_id')
        self.assertIsInstance(conversion[0].field, DynamicNdbKeyStringField)
        _type = conversion[0].field.get_type()
        self.assertIsInstance(_type, NdbKeyStringField)
        self.assertEqual(_type._type, String)

        self.assertEqual(conversion[1].name, 'user')
        self.assertIsInstance(conversion[1].field, DynamicNdbKeyReferenceField)
        _type = conversion[1].field.get_type()
        self.assertIsInstance(_type, NdbKeyReferenceField)
        self.assertEqual(_type._type, UserType)



from google.appengine.ext import ndb

__author__ = 'ekampf'


class Address(ndb.Model):
    address1 = ndb.StringProperty()
    address2 = ndb.StringProperty()
    city = ndb.StringProperty()


class PhoneNumber(ndb.Model):
    area = ndb.StringProperty()
    number = ndb.StringProperty()


class Reader(ndb.Model):
    name = ndb.StringProperty()
    email = ndb.StringProperty()
    is_alive = ndb.BooleanProperty(default=True)


class Author(ndb.Model):
    name = ndb.StringProperty()
    email = ndb.StringProperty()
    addresses = ndb.LocalStructuredProperty(Address, repeated=True)
    mobile = ndb.LocalStructuredProperty(PhoneNumber)


class Tag(ndb.Model):
    name = ndb.StringProperty()


class Comment(ndb.Model):
    created_at = ndb.DateTimeProperty(auto_now_add=True)
    updated_at = ndb.DateTimeProperty(auto_now=True)
    body = ndb.StringProperty()


class ArticleReader(ndb.Model):
    article_key = ndb.KeyProperty(kind='Article')
    reader_key = ndb.KeyProperty(kind='Reader')

    read_at = ndb.DateTimeProperty(auto_now_add=True)


class Article(ndb.Model):
    headline = ndb.StringProperty(required=True)
    summary = ndb.StringProperty()
    body = ndb.TextProperty()
    body_hash = ndb.ComputedProperty(lambda self: self.calc_body_hash())
    keywords = ndb.StringProperty(repeated=True)

    author_key = ndb.KeyProperty(kind='Author')
    tags = ndb.KeyProperty(Tag, repeated=True)

    created_at = ndb.DateTimeProperty(auto_now_add=True)
    updated_at = ndb.DateTimeProperty(auto_now=True)

    def calc_body_hash(self):
        import hashlib
        return hashlib.md5(self.body).hexdigit() if self.body else None

